from enum import Enum


class DataType(str, Enum):
    """Represents the different data types for a table column."""

    Bool = "BOOL"
    """32-bit IEEE 754 floating-point number."""

    Float32 = "FLOAT32"
    """32-bit IEEE 754 floating-point number."""

    Float64 = "FLOAT64"
    """64-bit IEEE 754 floating-point number."""

    Int32 = "INT32"
    """32-bit signed integers."""

    Int64 = "INT64"
    """64-bit signed integers."""

    String = "STRING"
    """Arbitrary string data."""

    Timestamp = "TIMESTAMP"
    """Date and time represented in UTC with millisecond precision."""



"""Implementation of DataFrameClient."""

from typing import List, Optional

from nisystemlink.clients import core
from nisystemlink.clients.core._uplink._base_client import BaseClient
from nisystemlink.clients.core._uplink._methods import (
    delete,
    get,
    patch,
    post,
    response_handler,
)
from nisystemlink.clients.core.helpers import IteratorFileLike
from requests.models import Response
from uplink import Body, Field, Path, Query

from . import models


class DataFrameClient(BaseClient):
    def __init__(self, configuration: Optional[core.HttpConfiguration] = None):
        """Initialize an instance.

        Args:
            configuration: Defines the web server to connect to and information about
                how to connect. If not provided, the
                :class:`HttpConfigurationManager <nisystemlink.clients.core.HttpConfigurationManager>`
                is used to obtain the configuration.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service.
        """
        if configuration is None:
            configuration = core.HttpConfigurationManager.get_configuration()

        super().__init__(configuration, "/nidataframe/v1/")

    @get("")
    def api_info(self) -> models.ApiInfo:
        """Get information about available API operations.

        Returns:
            Information about available API operations.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service.
        """
        ...

    @get(
        "tables",
        args=[
            Query("take"),
            Query("id"),
            Query("orderBy"),
            Query("orderByDescending"),
            Query("continuationToken"),
            Query("workspace"),
        ],
    )
    def list_tables(
        self,
        take: Optional[int] = None,
        id: Optional[List[str]] = None,
        order_by: Optional[models.OrderBy] = None,
        order_by_descending: Optional[bool] = None,
        continuation_token: Optional[str] = None,
        workspace: Optional[List[str]] = None,
    ) -> models.PagedTables:
        """Lists available tables on the SystemLink DataFrame service.

        Args:
            take: Limits the returned list to the specified number of results. Defaults to 1000.
            id: List of table IDs to filter by.
            order_by: The sort order of the returned list of tables.
            order_by_descending: Whether to sort descending instead of ascending. Defaults to false.
            continuation_token: The token used to paginate results.
            workspace: List of workspace IDs to filter by.

        Returns:
            The list of tables with a continuation token.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service
                or provided an invalid argument.
        """
        ...

    @post("tables", return_key="id")
    def create_table(self, table: models.CreateTableRequest) -> str:
        """Create a new table with the provided metadata and column definitions.

        Args:
            table: The request to create the table.

        Returns:
            The ID of the newly created table.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service
                or provided an invalid argument.
        """
        ...

    @post("query-tables")
    def query_tables(self, query: models.QueryTablesRequest) -> models.PagedTables:
        """Queries available tables on the SystemLink DataFrame service and returns their metadata.

        Args:
            query: The request to query tables.

        Returns:
            The list of tables with a continuation token.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service
                or provided an invalid argument.
        """
        ...

    @get("tables/{id}")
    def get_table_metadata(self, id: str) -> models.TableMetadata:
        """Retrieves the metadata and column information for a single table identified by its ID.

        Args:
            id (str): Unique ID of a data table.

        Returns:
            The metadata for the table.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service
                or provided an invalid argument.
        """
        ...

    @patch("tables/{id}", args=[Path, Body])
    def modify_table(self, id: str, update: models.ModifyTableRequest) -> None:
        """Modify properties of a table or its columns.

        Args:
            id: Unique ID of a data table.
            update: The metadata to update.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service
                or provided an invalid argument.
        """
        ...

    @delete("tables/{id}")
    def delete_table(self, id: str) -> None:
        """Deletes a table.

        Args:
            id (str): Unique ID of a data table.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service
                or provided an invalid argument.
        """
        ...

    @post("delete-tables", args=[Field("ids")])
    def delete_tables(
        self, ids: List[str]
    ) -> Optional[models.DeleteTablesPartialSuccess]:
        """Deletes multiple tables.

        Args:
            ids (List[str]): List of unique IDs of data tables.

        Returns:
            A partial success if any tables failed to delete, or None if all
            tables were deleted successfully.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service
                or provided an invalid argument.
        """
        ...

    @post("modify-tables")
    def modify_tables(
        self, updates: models.ModifyTablesRequest
    ) -> Optional[models.ModifyTablesPartialSuccess]:
        """Modify the properties associated with the tables identified by their IDs.

        Args:
            updates: The table modifications to apply.

        Returns:
            A partial success if any tables failed to be modified, or None if all
            tables were modified successfully.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service
                or provided an invalid argument.
        """
        ...

    @get(
        "tables/{id}/data",
        args=[
            Path("id"),
            Query("columns"),
            Query("orderBy"),
            Query("orderByDescending"),
            Query("take"),
            Query("continuationToken"),
        ],
    )
    def get_table_data(
        self,
        id: str,
        columns: Optional[List[str]] = None,
        order_by: Optional[List[str]] = None,
        order_by_descending: Optional[bool] = None,
        take: Optional[int] = None,
        continuation_token: Optional[str] = None,
    ) -> models.PagedTableRows:
        """Reads raw data from the table identified by its ID.

        Args:
            id: Unique ID of a data table.
            columns: Columns to include in the response. Data will be returned in the same order as
                the columns. If not specified, all columns are returned.
            order_by: List of columns to sort by. Multiple columns may be specified to order rows
                that have the same value for prior columns. The columns used for ordering do not
                need to be included in the columns list, in which case they are not returned. If
                not specified, then the order in which results are returned is undefined.
            order_by_descending: Whether to sort descending instead of ascending. Defaults to false.
            take: Limits the returned list to the specified number of results. Defaults to 500.
            continuation_token: The token used to paginate results.

        Returns:
            The table data and total number of rows with a continuation token.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service
                or provided an invalid argument.
        """
        ...

    @post("tables/{id}/data", args=[Path, Body])
    def append_table_data(self, id: str, data: models.AppendTableDataRequest) -> None:
        """Appends one or more rows of data to the table identified by its ID.

        Args:
            id: Unique ID of a data table.
            data: The rows of data to append and any additional options.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service
                or provided an invalid argument.
        """
        ...

    @post("tables/{id}/query-data", args=[Path, Body])
    def query_table_data(
        self, id: str, query: models.QueryTableDataRequest
    ) -> models.PagedTableRows:
        """Reads rows of data that match a filter from the table identified by its ID.

        Args:
            id: Unique ID of a data table.
            query: The filtering and sorting to apply when reading data.

        Returns:
            The table data and total number of rows with a continuation token.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service
                or provided an invalid argument.
        """
        ...

    @post("tables/{id}/query-decimated-data", args=[Path, Body])
    def query_decimated_data(
        self, id: str, query: models.QueryDecimatedDataRequest
    ) -> models.TableRows:
        """Reads decimated rows of data from the table identified by its ID.

        Args:
            id: Unique ID of a data table.
            query: The filtering and decimation options to apply when reading data.

        Returns:
            The decimated table data.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service
                or provided an invalid argument.
        """
        ...

    def _iter_content_filelike_wrapper(response: Response) -> IteratorFileLike:
        return IteratorFileLike(response.iter_content(chunk_size=4096))

    @response_handler(_iter_content_filelike_wrapper)
    @post("tables/{id}/export-data", args=[Path, Body])
    def export_table_data(
        self, id: str, query: models.ExportTableDataRequest
    ) -> IteratorFileLike:
        """Exports rows of data that match a filter from the table identified by its ID.

        Args:
            id: Unique ID of a data table.
            query: The filtering, sorting, and export format to apply when exporting data.

        Returns:
            A file-like object for reading the exported data.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service
                or provided an invalid argument.
        """
        ...



# -*- coding: utf-8 -*-

"""Implementation of TagValueWriter."""

import datetime
import typing
from typing import Any, Awaitable, Generic, Optional, TypeVar

from nisystemlink.clients import tag as tbase

_Any = TypeVar("_Any")


class TagValueWriter(Generic[_Any]):
    """Represents the ability to write a single tag's value using an :class:`ITagWriter`."""

    def __init__(self, writer: tbase.ITagWriter, tag: tbase.TagData) -> None:
        """Initialize an instance.

        Args:
            writer: The :class:`ITagWriter` to use when writing values.
            tag: The tag whose values will be written.
            path: The path of the tag whose values will be written.

        Raises:
            ValueError: if ``tag`` is not a tag of a valid type and with a valid path.
        """
        if tag.data_type == tbase.DataType.UNKNOWN:
            raise ValueError("tag.type cannot be UNKNOWN")

        self._path = tag.validate_path()
        self._data_type = tag.data_type
        self.__writer = writer

    @property
    def data_type(self) -> tbase.DataType:  # noqa: D401
        """The data type of the tag associated with the value."""
        return self._data_type

    @property
    def path(self) -> str:  # noqa: D401
        """The path of the tag associated with the value."""
        return self._path

    @property
    def _writer(self) -> tbase.ITagWriter:  # noqa: D401
        """The underlying :class:`ITagWriter` for writing values."""
        return self.__writer

    def write(
        self, value: _Any, *, timestamp: Optional[datetime.datetime] = None
    ) -> None:
        """Write the tag's value.

        Args:
            value: The tag value to write.
            timestamp: A custom timestamp to associate with the value, or None to have
                the server specify the timestamp.

        Raises:
            ReferenceError: if the underlying writer has been closed.
            ApiException: if the API call fails.
        """
        self._writer.write(
            self._path, self._data_type, typing.cast(Any, value), timestamp=timestamp
        )

    def write_async(
        self, value: _Any, *, timestamp: Optional[datetime.datetime] = None
    ) -> Awaitable[None]:
        """Write the tag's value.

        Args:
            value: The tag value to write.
            timestamp: A custom timestamp to associate with the value, or None to have
                the server specify the timestamp.

        Raises:
            ReferenceError: if the underlying writer has been closed.
            ApiException: if the API call fails.
        """
        return self._writer.write_async(
            self._path, self._data_type, typing.cast(Any, value), timestamp=timestamp
        )



from enum import Enum


class ColumnType(str, Enum):
    """Represents the different column types for a table column."""

    Normal = "NORMAL"
    """The column has no special properties. This is the default."""

    Index = "INDEX"
    """The column provides a unique value per row. Each table must provide
    exactly one INDEX column. The column's :class:`.DataType` must be INT32,
    INT64, or TIMESTAMP."""

    Nullable = "NULLABLE"
    """Rows may contain null values for this column. When appending rows,
    NULLABLE columns may be left out entirely, in which case all rows being
    appended will use null values for that column."""



"""Wrappers around uplink HTTP decorators with proper type annotations."""

from typing import Any, Callable, Optional, Sequence, Tuple, TypeVar, Union

from uplink import (
    Body,
    commands,
    json,
    response_handler as uplink_response_handler,
    returns,
)

F = TypeVar("F", bound=Callable[..., Any])


def get(path: str, args: Optional[Sequence[Any]] = None) -> Callable[[F], F]:
    """Annotation for a GET request."""

    def decorator(func: F) -> F:
        return commands.get(path, args=args)(func)  # type: ignore

    return decorator


def post(
    path: str,
    args: Optional[Sequence[Any]] = None,
    return_key: Optional[Union[str, Tuple[str, ...]]] = None,
) -> Callable[[F], F]:
    """Annotation for a POST request with a JSON request body. If args is not
    specified, defaults to a single argument that represents the request body.
    """

    def decorator(func: F) -> F:
        result = json(commands.post(path, args=args or (Body,))(func))
        if return_key:
            result = returns.json(key=return_key)(result)
        return result  # type: ignore

    return decorator


def patch(path: str, args: Optional[Sequence[Any]] = None) -> Callable[[F], F]:
    """Annotation for a PATCH request with a JSON request body."""

    def decorator(func: F) -> F:
        return json(commands.patch(path, args=args)(func))  # type: ignore

    return decorator


def delete(path: str, args: Optional[Sequence[Any]] = None) -> Callable[[F], F]:
    """Annotation for a DELETE request."""

    def decorator(func: F) -> F:
        return commands.delete(path, args=args)(func)  # type: ignore

    return decorator


def response_handler(
    handler: Any, requires_consumer: Optional[bool] = False
) -> Callable[[F], F]:
    """Annotation for creating custom response handlers."""

    def decorator(func: F) -> F:
        return uplink_response_handler(handler, requires_consumer)(func)  # type: ignore

    return decorator



from nisystemlink.clients.core._uplink._json_model import JsonModel


class UploadArtifactResponse(JsonModel):
    """Response for an artifact upload request."""

    id: str
    """Information about the uploaded artifact."""



from typing import List


from nisystemlink.clients.core import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel


class DeleteSpecificationsPartialSuccess(JsonModel):
    """The results of deleting multiple specs when one or more of the specs could not be deleted."""

    deleted_spec_ids: List[str]
    """Global IDs of the deleted specifications."""

    failed_spec_ids: List[str]
    """Global IDs of the specifications that could not be deleted."""

    error: ApiError



# -*- coding: utf-8 -*-

from ._data_type import DataType
from ._retention_type import RetentionType
from ._tag_data import TagData
from ._tag_with_aggregates import TagWithAggregates
from ._async_tag_query_result_collection import AsyncTagQueryResultCollection
from ._itag_reader import ITagReader
from ._itag_writer import ITagWriter
from ._buffered_tag_writer import BufferedTagWriter
from ._tag_value_reader import TagValueReader
from ._tag_value_writer import TagValueWriter
from ._tag_update_fields import TagUpdateFields
from ._tag_data_update import TagDataUpdate
from ._tag_path_utilities import TagPathUtilities
from ._tag_query_result_collection import TagQueryResultCollection
from ._tag_subscription import TagSubscription
from ._tag_selection import TagSelection
from ._tag_manager import TagManager

# flake8: noqa



from typing import Dict, List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._column import Column


class CreateTableRequest(JsonModel):
    """Contains information needed to create a table, including its properties and column definitions."""

    columns: List[Column]
    """The list of columns in the table. Exactly one column must have a :class:`.ColumnType` of INDEX."""

    name: Optional[str] = None
    """The name to associate with the table. When not specified, a name will be
    assigned from the table's ID."""

    properties: Optional[Dict[str, str]] = None
    """User-defined properties to associate with the table."""

    workspace: Optional[str] = None
    """The workspace to create the table in. Uses the default workspace when not specified."""



from typing import Dict, List, Optional

from nisystemlink.clients.core._uplink._json_model import JsonModel


class ColumnMetadataPatch(JsonModel):
    """Specifies column properties to add, modify, or delete when editing table metadata."""

    name: str
    """The name of the column to modify."""

    properties: Dict[str, Optional[str]]
    """The properties to modify. A map of key value properties containing the metadata
    to be added or modified. Setting a property value to ``None`` will delete the
    property."""


class ModifyTableRequest(JsonModel):
    """Contains the metadata properties to modify. Values not included will remain unchanged."""

    metadata_revision: Optional[int] = None
    """When specified, this is an integer that must match the last known
    revision number of the table, incremented by one. If it doesn't match the
    current ``metadataRevision`` incremented by one at the time of execution, the
    modify request will be rejected with a 409 Conflict. This is used to ensure
    that changes to this table's metadata are based on a known, previous
    state."""

    name: Optional[str] = None
    """The new name of the table. Setting to ``None`` will reset the name to the table's ID."""

    workspace: Optional[str] = None
    """The new workspace for the table. Setting to ``None`` will reset to the
    default workspace. Changing the workspace requires permission to delete the
    table in its current workspace and permission to create the table in its new
    workspace."""

    properties: Optional[Dict[str, Optional[str]]] = None
    """The properties to modify. A map of key value properties containing the
    metadata to be added or modified. Setting a property value to ``None`` will
    delete the property."""

    columns: Optional[List[ColumnMetadataPatch]] = None
    """Updates to the column properties. Cannot add or remove columns, or change the name of a column."""



"""Implementation of ArtifactClient"""

from typing import BinaryIO, Optional

from nisystemlink.clients import core
from nisystemlink.clients.core._uplink._base_client import BaseClient
from nisystemlink.clients.core._uplink._methods import get, post, response_handler
from nisystemlink.clients.core.helpers._iterator_file_like import IteratorFileLike
from requests.models import Response
from uplink import Part, Path

from . import models


class ArtifactClient(BaseClient):
    def __init__(self, configuration: Optional[core.HttpConfiguration] = None):
        """Initialize an instance.

        Args:
            configuration: Defines the web server to connect to and information about
                how to connect. If not provided, the
                :class:`HttpConfigurationManager <nisystemlink.clients.core.HttpConfigurationManager>`
                is used to obtain the configuration.

        Raises:
            ApiException: if unable to communicate with the  Notebook execution Service.
        """
        if configuration is None:
            configuration = core.HttpConfigurationManager.get_configuration()

        super().__init__(configuration, base_path="/ninbartifact/v1/")

    @post("artifacts")
    def __upload_artifact(
        self, workspace: Part, artifact: Part
    ) -> models.UploadArtifactResponse:
        """Uploads an artifact  using multipart/form-data headers to send the file payload in the HTTP body.

        Args:
            workspace: The workspace containing the artifact.
            artifact: The artifact to upload.

        Returns:
            UploadArtifactResponse: The response containing the artifact ID.

        """

    def upload_artifact(
        self, workspace: str, artifact: BinaryIO
    ) -> models.UploadArtifactResponse:
        """Uploads an artifact.

        Args:
            workspace: The workspace containing the artifact.
            artifact: The artifact to upload.

        Returns:
            UploadArtifactResponse: The response containing the artifact ID.

        """
        response = self.__upload_artifact(
            workspace=workspace,
            artifact=artifact,
        )

        return response

    def _iter_content_filelike_wrapper(response: Response) -> IteratorFileLike:
        return IteratorFileLike(response.iter_content(chunk_size=4096))

    @response_handler(_iter_content_filelike_wrapper)
    @get("artifacts/{id}")
    def download_artifact(self, id: Path) -> IteratorFileLike:
        """Downloads an artifact.

        Args:
            id: The ID of the artifact to download.

        Returns:
            A file-like object for reading the artifact content.

        """
        ...



# -*- coding: utf-8 -*-

"""Implementation of TimestampUtilities."""

import datetime

from typing_extensions import final


@final
class TimestampUtilities:
    """Provides utilities for reading and writing timestamps as strings.

    Clients do not typically need to call methods on this class.
    """

    def __init_subclass__(cls) -> None:
        raise TypeError("type 'TimestampUtilities' is not an acceptable base type")

    def __init__(self) -> None:
        raise TypeError("Can't instantiate static class 'TimestampUtilities'")

    @classmethod
    def datetime_to_str(cls, value: datetime.datetime) -> str:
        """Convert the given ``datetime.datetime`` into a string timestamp in the standard format used in SystemLink.

        Args:
            value: The date and time to convert.

        Returns:
            The string representation of the timestamp.
        """
        return datetime.datetime.utcfromtimestamp(value.timestamp()).isoformat() + "Z"

    @classmethod
    def str_to_datetime(cls, timestamp: str) -> datetime.datetime:
        """Attempt to parse a SystemLink-formatted timestamp string into a ``datetime.datetime``.

        Args:
            timestamp: The timestamp to parse, in the standard format used in
                SystemLink.

        Returns:
            The parsed datetime.

        Raises:
            ValueError: if the timestamp format is not as expected
        """
        # Python's supported ISO format requires exactly 6 digits after the
        # decimal, and doesn't support "Z" as the timezone
        # Valid format is: YYYY-MM-DDThh:mm:ss.ssssss+NN:NN
        if not timestamp.endswith("Z"):
            raise ValueError(
                "Given timestamp doesn't end with 'Z': '{}'".format(timestamp)
            )
        timestamp = timestamp[:-1].ljust(26, "0")[:26] + "+0000"
        # Note to users: this will be in UTC time; to get a local datetime, you
        # can use value.astimezone()
        return datetime.datetime.strptime(timestamp, "%Y-%m-%dT%H:%M:%S.%f%z")



from datetime import datetime
from typing import List, Optional, Union

from nisystemlink.clients.core._uplink._with_paging import WithPaging
from pydantic import StrictBool, StrictInt

from ._order_by import OrderBy


class QueryTablesRequest(WithPaging):
    """Request parameters for querying tables."""

    filter: str
    """The table query filter in `Dynamic LINQ`_ format.

    .. _Dynamic LINQ: https://github.com/ni/systemlink-OpenAPI-documents/wiki/Dynamic-Linq-Query-Language

    Allowed properties in the filter are:

    * ``createdAt``: DateTime the table was created
    * ``createdWithin``: TimeSpan in which the table was created
    * ``id``: String value uniquely identifying the table
    * ``name``: String name for the table
    * ``metadataModifiedAt``: DateTime the table's metadata was last modified
    * ``metadataModifiedWithin``: TimeSpan in which the table's metadata was
      last modified
    * ``properties``: Dictionary with string keys and values representing table
      metadata
    * ``rowsModifiedAt``: DateTime rows were last appended to the table
    * ``rowsModifiedWithin``: TimeSpan within rows were last appended to the
      table
    * ``rowCount``: Int32 number of rows in the table
    * ``supportsAppend``: Boolean indicating whether or not the table supports
      appending additional rows of data
    * ``workspace``: String value ID of the workspace the table belongs to
    * ``workspaceName``: String value name of the workspace the table belongs to

    Allowed constants in the filter are:

    * ``RelativeTime.CurrentDay``: TimeSpan representing the elapsed time
      between now and the start of the current day
    * ``RelativeTime.CurrentWeek``: TimeSpan representing the elapsed time
      between now and the start of the current week
    * ``RelativeTime.CurrentMonth``: TimeSpan representing the elapsed time
      between now and the start of the current month
    * ``RelativeTime.CurrentYear``: TimeSpan representing the elapsed time
      between now and the start of the current year
    """

    substitutions: Optional[List[Union[StrictInt, StrictBool, str, None]]] = None
    """Make substitutions in the query filter expression.

    Substitutions for the query expression are indicated by non-negative
    integers that are prefixed with the ``@`` symbol. Each substitution in the given
    expression will be replaced by the element at the corresponding index
    (zero-based) in this list. For example, ``@0`` in the filter expression will be
    replaced with the element at the zeroth index of the substitutions list.
    """

    reference_time: Optional[datetime] = None
    """The date and time to use as the reference point for `RelativeTime` filters,
    including time zone information. Defaults to the time on the server in UTC."""

    take: Optional[int] = None
    """Limits the returned list to the specified number of results."""

    order_by: Optional[OrderBy] = None
    """The sort order of the returned list of tables."""

    order_by_descending: Optional[bool] = None
    """Whether to sort descending instead of ascending.

    The elements in the list are sorted ascending by default. If the
    orderByDescending parameter is specified, the elements in the list are
    sorted based on it's value. The orderByDescending value must be a boolean
    string. The elements in the list are sorted ascending if false and
    descending if true.
    """



# -*- coding: utf-8 -*-

"""Implementation of ApiError."""

from typing import List, Optional

from ._uplink._json_model import JsonModel


class ApiError(JsonModel):
    """Represents the standard error structure for SystemLink API responses."""

    name: Optional[str] = None
    """String error code."""

    code: Optional[int] = None
    """Numeric error code."""

    message: Optional[str] = None
    """Complete error message."""

    args: List[str] = []
    """Positional arguments for the error code."""

    resource_type: Optional[str] = None
    """Type of resource associated with the error."""

    resource_id: Optional[str] = None
    """Identifier of the resource associated with the error."""

    inner_errors: List["ApiError"] = []
    """Inner errors when the top-level error represents more than one error."""



# -*- coding: utf-8 -*-

"""Implementation of JupyterHttpConfiguration."""

import os

from nisystemlink.clients import core


class JupyterHttpConfiguration(core.HttpConfiguration):
    """An :class:`HttpConfiguration` for Jupyter notebooks running in a SystemLink environment."""

    _HTTP_URI_ENV_VAR = "SYSTEMLINK_HTTP_URI"
    _HTTP_API_KEY_ENV_VAR = "SYSTEMLINK_API_KEY"

    def __init__(self) -> None:
        """Initialize a configuration for SystemLink using API key-based
        authentication provided through environment variables.

        Raises:
            KeyError: if the expected environment variables are not set.
        """
        super().__init__(
            os.environ[self._HTTP_URI_ENV_VAR], os.environ[self._HTTP_API_KEY_ENV_VAR]
        )



# -*- coding: utf-8 -*-

"""Implementation of TagPathUtilities."""

from typing_extensions import final


@final
class TagPathUtilities:
    """Contains helper methods for interacting with tag paths."""

    def __init_subclass__(cls) -> None:
        raise TypeError("type 'TagPathUtilities' is not an acceptable base type")

    def __init__(self) -> None:
        raise TypeError("Can't instantiate static class 'TagPathUtilities'")

    @classmethod
    def validate(cls, path: str) -> str:
        """Validate ``path`` as an input tag path.

        Clients do not typically need to call this method directly.

        Args:
            path: The tag path to validate.

        Returns:
            The validated path.

        Raises:
            ValueError: if the path is invalid.
            ValueError: if ``path`` is None.
        """
        if path is None or path.lstrip() == "":
            raise ValueError("Tag path cannot be None or empty")

        if "*" in path:
            raise ValueError("Tag path cannot contain a *")

        return path

    @classmethod
    def validate_query(cls, path: str) -> str:
        """Validate ``path`` as a tag path query.

        Clients do not typically need to call this method directly.

        Args:
            path: The tag path to validate.

        Returns:
            The validated path.

        Raises:
            ValueError: if the path is invalid.
            ValueError: if ``path`` is None.
        """
        if path is None or path.lstrip() == "":
            raise ValueError("Tag path cannot be None or empty")

        if "," in path:
            raise ValueError("Tag path cannot contain a ,")

        return path



from typing import Literal

# TODO: Migrate to Enum when this change is released: https://github.com/prkumar/uplink/pull/282
OrderBy = Literal[
    "CREATED_AT", "METADATA_MODIFIED_AT", "NAME", "NUMBER_OF_ROWS", "ROWS_MODIFIED_AT"
]
"""Possible options for sorting when querying tables.

* ``CREATED_AT``: The date and time the table was created.
* ``METADATA_MODIFIED_AT``: The date and time the table's metadata properties were modified.
* ``NAME``: The name of the table.
* ``NUMBER_OF_ROWS``: The number of rows of data in the table.
* ``ROWS_MODIFIED_AT``: Date and time rows were most recently appended to the table.
"""



# mypy: disable-error-code = misc

from json import loads
from typing import Any, Callable, Dict, get_origin, Optional, Type, Union

from nisystemlink.clients import core
from pydantic import parse_obj_as
from requests import JSONDecodeError, Response
from uplink import commands, Consumer, converters, response_handler, utils

from ._json_model import JsonModel


@response_handler
def _handle_http_status(response: Response) -> Optional[Response]:
    """Checks an HTTP response's status code and raises an exception if necessary."""
    if 200 <= response.status_code < 300:
        # Return None for "204 No Content" responses.
        if response.status_code == 204:
            return None
        return response

    msg = "Server responded with <{} {}> ({}).".format(
        response.status_code, response.reason, response.url
    )

    try:
        content = response.json()
        if content and "error" in content:
            err_obj = core.ApiError.parse_obj(content["error"])
        else:
            err_obj = None

        raise core.ApiException(
            msg, error=err_obj, http_status_code=response.status_code
        )
    except JSONDecodeError:
        if response.text:
            msg += ":\n\n" + response.text
        raise core.ApiException(msg, http_status_code=response.status_code)


class _JsonModelConverter(converters.Factory):
    def create_request_body_converter(
        self, _class: Type, _: commands.RequestDefinition
    ) -> Optional[Callable[[JsonModel], Dict]]:
        def encoder(model: JsonModel) -> Dict:
            return loads(model.json(by_alias=True, exclude_unset=True))

        if utils.is_subclass(_class, JsonModel):
            return encoder
        else:
            return None

    def create_response_body_converter(
        self, _class: Type, _: commands.RequestDefinition
    ) -> Optional[Callable[[Response], Any]]:
        def decoder(response: Response) -> Any:
            try:
                data = response.json()
            except AttributeError:
                data = response

            return parse_obj_as(_class, data)

        if get_origin(_class) is Union or utils.is_subclass(_class, JsonModel):
            return decoder
        else:
            return None


class BaseClient(Consumer):
    """Base class for SystemLink clients, built on top of `Uplink <https://github.com/prkumar/uplink>`_."""

    def __init__(self, configuration: core.HttpConfiguration, base_path: str = ""):
        """Initialize an instance.

        Args:
            configuration: Defines the web server to connect to and information about how to connect.
            base_path: The base path for all API calls.
        """
        super().__init__(
            base_url=configuration.server_uri + base_path,
            converter=_JsonModelConverter(),
            hooks=[_handle_http_status],
        )
        if configuration.api_keys:
            self.session.headers.update(configuration.api_keys)



# -*- coding: utf-8 -*-

"""Implementation of RetentionType."""

import enum


class RetentionType(enum.Enum):
    """Represents the different ways for the SystemLink tag historian to retain the history of tag values."""

    INVALID = 0
    """An unknown or invalid tag retention type.

    Not a valid input to API calls, but used to represent tags whose retention type
    isn't recognized.
    """

    NONE = 1
    """No history for the tag's value is retained."""

    DURATION = 2
    """Historical values for the tag are retained for a set number of days."""

    COUNT = 3
    """A set number of historical values for the tag are retained."""

    PERMANENT = 4
    """All of the tag's values are retained until the tag is deleted."""



from typing import List, Optional

from nisystemlink.clients.core import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel
from nisystemlink.clients.spec.models._specification import (
    SpecificationCreation,
    SpecificationDefinition,
    SpecificationServerManaged,
    SpecificationUserManaged,
)


class CreateSpecificationsRequest(JsonModel):
    """Create multiple specifications."""

    specs: Optional[List[SpecificationDefinition]] = None
    """List of specifications to be created."""


class CreatedSpecification(
    SpecificationServerManaged, SpecificationUserManaged, SpecificationCreation
):
    """A specification successfully created on the server."""


class CreateSpecificationsPartialSuccess(JsonModel):
    """When some specs can not be created, this contains the list that was and was not created."""

    created_specs: Optional[List[CreatedSpecification]] = None
    """Information about the created specification(s)"""

    failed_specs: Optional[List[SpecificationDefinition]] = None
    """List of specification requests that failed during creation."""

    error: Optional[ApiError] = None



# -*- coding: utf-8 -*-

"""Implementation of TagQueryResultCollection."""

import abc
from typing import Iterable, List, Optional

from nisystemlink.clients import core, tag as tbase


class TagQueryResultCollection(abc.ABC):
    """Represents a paginated list of tags returned by a query.

    Iterating over the collection makes additional server requests to retrieve pages
    after the first.
    """

    def __init__(
        self, first_page: List[tbase.TagData], total_count: int, skip: int
    ) -> None:
        """Initialize an instance with the first page of query results.

        Args:
            first_page: The first page of results, or None if there are no results.
            total_count: The total number of results in the query.
            skip: The skip used for the first page of results.
        """
        self._first_page = None  # type: Optional[List[tbase.TagData]]
        if first_page:
            if skip >= total_count:
                raise core.ApiException(
                    "skip is >= totalCount, but the tag list isn't empty"
                )
            self._first_page = first_page
        else:
            pass  # leave it as None, even if passed in as []
        self._use_cached_page = True
        self._total_count = total_count
        self._skip = skip

    @property
    def total_count(self) -> int:  # noqa: D401
        """The total number of tags matched by the query at the time the query was made."""
        return self._total_count

    def __iter__(self) -> Iterable[List[tbase.TagData]]:
        """Enumerate over the pages of tag query results.

        Calls to ``next(iter())`` may throw :class:`.ApiException`.

        Returns:
            The created enumerator.
        """
        skip = self._skip

        if self._use_cached_page:
            self._use_cached_page = False
            page = self._first_page
            self._first_page = None

            if not page:
                return

            skip += len(page)
            yield page

            if skip >= self._total_count:
                return

        while True:
            page = self._query_page(skip)

            if not page:
                return

            skip += len(page)
            yield page

            if skip >= self._total_count:
                break

    @abc.abstractmethod
    def _query_page(self, skip: int) -> List[tbase.TagData]:
        """Query for a single page of results and updates :attr:`total_count`.

        Args:
            skip: The skip to use in the query.

        Returns:
            The page of results, or None if there are no more results.

        Raises:
            ApiException: if the API call fails.
        """
        ...



"""Implementation of SpecClient"""

from typing import List, Optional

from nisystemlink.clients import core
from nisystemlink.clients.core._uplink._base_client import BaseClient
from nisystemlink.clients.core._uplink._methods import get, post
from uplink import Field

from . import models


class SpecClient(BaseClient):
    def __init__(self, configuration: Optional[core.HttpConfiguration] = None):
        """Initialize an instance.

        Args:
            configuration: Defines the web server to connect to and information about
                how to connect. If not provided, the
                :class:`HttpConfigurationManager <nisystemlink.clients.core.HttpConfigurationManager>`
                is used to obtain the configuration.

        Raises:
            ApiException: if unable to communicate with the Spec Service.
        """
        if configuration is None:
            configuration = core.HttpConfigurationManager.get_configuration()

        super().__init__(configuration, base_path="/nispec/v1/")

    @get("")
    def api_info(self) -> models.V1Operations:
        """Get information about available API operations.

        Returns:
            Information about available API operations.

        Raises:
            ApiException: if unable to communicate with the DataFrame Service.
        """
        ...

    @post("specs")
    def create_specs(
        self, specs: models.CreateSpecificationsRequest
    ) -> models.CreateSpecificationsPartialSuccess:
        """Creates one or more specifications.

        Args:
            specs: A list of specifications to create.

        Returns:
            A list of specs that were successfully created and ones that failed to be created.

        Raises:
            ApiException: if unable to communicate with the `/nispec` service or if there are
            invalid arguments.
        """
        ...

    @post("delete-specs", args=[Field("ids")])
    def delete_specs(
        self, ids: List[str]
    ) -> Optional[models.DeleteSpecificationsPartialSuccess]:
        """Deletes one or more specifications by global id.

        Args:
            ids: a list of specification ids. Note that these are the global ids and not the
            `specId` that is local to a product and workspace.

        Returns:
            None if all deletes succeed otherwise a list of which ids failed and which succeeded.

        Raises:
            ApiException: if unable to communicate with the `nispec` service or if there are invalid
            arguments.
        """
        ...

    @post("query-specs")
    def query_specs(
        self, query: models.QuerySpecificationsRequest
    ) -> models.QuerySpecifications:
        """Queries for specs that match the filters.

        Args:
            query: The query contains a product id as well as a filter for specs under that product.

        Returns:
            A list of specifications that match the filter.
        """
        ...

    @post("update-specs")
    def update_specs(
        self, specs: models.UpdateSpecificationsRequest
    ) -> Optional[models.UpdateSpecificationsPartialSuccess]:
        """Updates one or more specifications.

        Update requires that the version field matches the version being updated from.

        Args:
            specs: a list of specifications that are to be updated. Must include the global id and
            each spec being updated must match the version currently on the server.

        Returns
            A list of specs that were successfully updated and a list of ones that were not along
            with error messages for updates that failed.
        """
        ...



from typing import List, Optional

from nisystemlink.clients.core import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel
from nisystemlink.clients.spec.models._specification import (
    Specification,
    SpecificationServerManaged,
    SpecificationUpdated,
    SpecificationUserManaged,
)


class UpdateSpecificationsRequest(JsonModel):

    specs: Optional[List[Specification]] = None
    """List of specifications to be updated."""


class UpdatedSpecification(
    SpecificationUserManaged,
    SpecificationServerManaged,
    SpecificationUpdated,
):
    """A specification that was updated on the server."""


class UpdateSpecificationsPartialSuccess(JsonModel):

    updated_specs: Optional[List[UpdatedSpecification]] = None
    """Information about each of the updated specification(s)."""

    failed_specs: Optional[List[Specification]] = None
    """Information about each of the specification request(s) that failed during the update."""

    error: Optional[ApiError] = None
    """Any errors that occurred."""



from typing import List

from nisystemlink.clients.core import ApiError
from nisystemlink.clients.core._uplink._json_model import JsonModel

from ._modify_tables_request import TableMetadataModification


class ModifyTablesPartialSuccess(JsonModel):
    """The result of modifying multiple tables when one or more tables could not be modified."""

    modified_table_ids: List[str]
    """The IDs of the tables that were successfully modified."""

    failed_modifications: List[TableMetadataModification]
    """The requested modifications that could not be applied."""

    error: ApiError
    """The error that occurred when modifying the tables."""



# -*- coding: utf-8 -*-

from ._api_error import ApiError
from ._api_exception import ApiException
from ._http_configuration import HttpConfiguration
from ._cloud_http_configuration import CloudHttpConfiguration
from ._jupyter_http_configuration import JupyterHttpConfiguration
from ._http_configuration_manager import HttpConfigurationManager

# flake8: noqa
